#!/usr/bin/python

# GDB debugging support
#
# Copyright (c) 2015 Linaro Ltd
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see
# <http://www.gnu.org/licenses/gpl-2.0.html>
#

# We don't need to do anything in our init file currently.

"""
Support routines for debugging QEMU under GDB
"""

__license__    = "GPL version 2 or (at your option) any later version"
