#include "stdbool.h"            /* bool (in exec/gdbstub.h) */
#include "stddef.h"             /* NULL */
#include "exec/gdbstub.h"       /* xml_builtin */

const char *const xml_builtin[][2] = {
  { NULL, NULL }
};
