/*
 * ARM Integrator Board Debug, switch and LED section
 *
 * Browse the data sheet:
 *
 *    http://infocenter.arm.com/help/index.jsp?topic=/com.arm.doc.dui0159b/Babbfijf.html
 *
 * Copyright (c) 2013 Alex Bennée <alex@bennee.com>
 *
 * This work is licensed under the terms of the GNU GPL, version 2 or later.
 * See the COPYING file in the top-level directory.
 */
#ifndef QEMU_INTEGRATOR_DEBUG_H
#define QEMU_INTEGRATOR_DEBUG_H

#define TYPE_INTEGRATOR_DEBUG "integrator_debug"

#endif
